import { EventEmitter, Injectable } from '@angular/core';
import { HttpClient } from '@angular/common/http';
import { Observable, tap } from 'rxjs';
import { ApplicationConfigService } from 'app/core/config/application-config.service';
import { ReviewRequest } from './reviewRequest.model';
import { Review, ReviewStatisticsDTO } from './review.model';
import { ReviewRating } from './reviewRating.model';

@Injectable({ providedIn: 'root' })
export class ReviewManagementService {
  resourceUrl = this.applicationConfigService.getEndpointFor('api/review');

  constructor(private http: HttpClient, private applicationConfigService: ApplicationConfigService) {}

  create(reviewRequest: ReviewRequest): Observable<ReviewRequest> {
    return this.http.post<ReviewRequest>(this.resourceUrl + '/create', reviewRequest).pipe(tap(() => this.resetStatistics()));
  }

  update(review: Review): Observable<Review> {
    return this.http.put<Review>(this.resourceUrl, review).pipe(tap(() => this.resetStatistics()));
  }

  delete(id: number): Observable<{}> {
    return this.http.delete(`${this.resourceUrl}/${id}`).pipe(tap(() => this.resetStatistics()));
  }

  find(id: number): Observable<Review> {
    return this.http.get<Review>(`${this.resourceUrl}/${id}`);
  }

  getAll(): Observable<Review[]> {
    return this.http.get<Review[]>(this.resourceUrl + '/all');
  }

  getAllWithStatistics(): Observable<Review[]> {
    return this.http.get<Review[]>(this.resourceUrl + '/allWithStatistics');
  }

  getAllByUser(): Observable<Review[]> {
    return this.http.get<Review[]>(this.resourceUrl + '/allByUser');
  }

  updateByUser(review: Review): Observable<Review> {
    return this.http.put<Review>(this.resourceUrl + '/updateByUser', review).pipe(tap(() => this.resetStatistics()));
  }

  notifyReviewer(review: Review): Observable<Review> {
    return this.http.post<Review>(this.resourceUrl + '/notifyReviewers', review);
  }

  rewardBadge(review: Review): Observable<Review> {
    return this.http.post<Review>(this.resourceUrl + '/rewardBadge', review).pipe(tap(() => this.resetStatistics()));
  }

  private reviewStatisticsCache: ReviewStatisticsDTO | undefined;

  reviewStatisticsCacheChanged = new EventEmitter<ReviewStatisticsDTO>();

  public getReviewStatistics(): Observable<ReviewStatisticsDTO> {
    if (this.reviewStatisticsCache) {
      return new Observable<ReviewStatisticsDTO>(subscriber => subscriber.next(this.reviewStatisticsCache)); // return cached value
    } else {
      // no cached value  ->  fetch from server and cache it  ->  return it to subscriber (component) via observable  ->  component updates its view accordingly
      return this.countAllByUser().pipe(
        // fetch from server
        tap(statistics => {
          this.reviewStatisticsCache = statistics;
          this.reviewStatisticsCacheChanged.emit(statistics);
        }) // cache it
      );
    }
  }

  resetStatistics() {
    this.reviewStatisticsCache = undefined;
    this.getReviewStatistics();
  }

  countAllByUser(): Observable<ReviewStatisticsDTO> {
    return this.http.post<ReviewStatisticsDTO>(this.resourceUrl + '/countAllByUser', null);
  }

  getAllReviewRatings(): Observable<ReviewRating[]> {
    return this.http.get<ReviewRating[]>(this.resourceUrl + '/reviewRatings');
  }
}
